/*----------------------------------------------------------------------+
|									|
|  Copyright (2017) LA Solutions Ltd, All rights reserved.		|
|  Copyright (1992) Bentley Systems, Inc., All rights reserved.		|
|									|
|  "MicroStationAPI", "MicroStation" and "MDL" are trademarks of Bentley	|
|  Systems, Inc.							|
|									|
|  Limited permission is hereby granted to reproduce and modify this	|
|  copyrighted material provided that the resulting code is used only 	|
|  in conjunction with Bentley Systems products under the terms of the	|
|  license agreement provided therein, and that this notice is retained	|
|  in its entirety in any such reproduction or modification.		|
|									|
+----------------------------------------------------------------------*/
/* ---------------------------------------------------------------------+
|																		|
|	BoxDependency														|
|																		|
|	The original BoxDependency creates a box (MicroStation shape elm)	|
|	that is dependent on a selection set of elements.  The box is drawn	|
|	as the minumim bounding box of the elements' range.  If one of the	|
|	original elements is moved, the box is recomputed and redrawn.		|
|																		|
|	keyins: BOX DEPEND CREATE [there must be a one or					|
|			elements in a selection set for this to succeed]			|
|			BOX DEPEND DROP testdepend_commandDropBox					|
|																		|
|	The OffsetDependency code creates a cell that is dependent on a		|
|	root element.  It is initially placed at a location determined by	|
|	a user <datapoint>.  When the cell is placed, a DependencyLinkage	|
|	is added that records the offset from the root element.  			|
|	If subsequently the root element is moved, the dependent cell is	|
|	also moved, to a position determined by the stored offset.  If the	|
|	dependent cell is moved, a new offset from the root element is		|
|	computed and stored in the DependencyLinkage.  The result is the	|
|	same as the behaviour of tag elements attached to a host element.	|
|																		|
|	keyins: BOX DEPEND OFFSET testdepend_commandOffset						|
|			[there must be at least one element in a selection set]		|
|																		|
+--------------------------------------------------------------------- */

/*----------------------------------------------------------------------+
|									|
|   Header File Dependencies						|
|									|
+----------------------------------------------------------------------*/
#include    <Mstn/MdlApi/MdlApi.h>
#include	<DgnView/SelectionSetManager.h>
#include    <Mstn/cmdlist.r.h>
#include    <string.h>
#include	<MicroStationUtilities.h>
#include 	<ModulVer.h>
#include	"BoxDependency.h"
#include	"BoxDependencyCallback.h"
#include	"BoxDependencyCmds.h"
#include	"BoxDependencyIds.h"
#include	"PlaceCellTool.h"
#include 	"MessageTracer.h"

USING_NAMESPACE_BENTLEY_DGNPLATFORM;

//	CONNECT MicroStation API
static BoxDependencyCallback	s_dependencyCallback;

///////////////////////////////////////////////////////////////////////
bool						IsSelectionSetActive ()
{
	SelectionSetManagerR	selection	{ SelectionSetManager::GetManager ()	};
	const bool				Active		{ selection.IsActive () };
    if (!Active)
    {
		WCharCP		msg	{ L"Requires selection set" };
		mdlOutput_messageCenter (DgnPlatform::OutputMessagePriority::Warning, msg, msg, DgnPlatform::OutputMessageAlert::Balloon);
	}
	return Active;
}
///////////////////////////////////////////////////////////////////////
bool				GetElementOriginById
(
DPoint3d&			origin,
ElementId			id,
DgnModelRefP		model
)
{
	bool			rc				{ false };
	ElementHandle	eh				(id, model);
	DisplayHandlerP	handler	= eh.GetDisplayHandler ();
	handler->GetSnapOrigin (eh, origin);	//	See also GetTransformOrigin()
	rc	= true;
	return rc;
}
/*----------------------------------------------------------------------+*//**
* !!!Describe Function Completely!!!								    	    *
* @bsimethod findHighLow   						*					    	*
* @param 	low     The low corner information	    	*
* @param 	high     The high corner information	    	*
* @param 	range     The range block from the contained element	    	*
* @return 	void			    	*
*								        *
* Author:   MarkAnderson 				02/01 		*
*								        *
+----------------------------------------------------------------------*/
void 				FindHighLow
(
DPoint3d&			low,
DPoint3d&			high,
DRange3d const&		range
)
{
    //handle the low end
    if  (range.low.x < low.x)
		low.x = range.low.x;
    if  (range.low.y < low.y)
		low.y = range.low.y;
    if  (range.low.z <low.z)
		low.z = range.low.z;
    //handle the high end
    if  (range.high.x > high.x)
    	high.x = range.high.x;
    if  (range.high.y > high.y)
    	high.y = range.high.y;
    if  (range.high.z > high.z)
    	high.z = range.high.z;
}
///////////////////////////////////////////////////////////////////////
void 					MakeShapeBoundary
(
DPoint3d*				range,
DPoint3d const&			low,
DPoint3d const&			high
)
{
    range [0]	= low;
    range [4]	= low;

    range [2].x = high.x;
    range [2].y = high.y;
    range [2].z = high.z;

    range [1].x = high.x;
    range [1].y = low.y;
    range [1].z = (low.z + high.z) / 2;

    range [3].x = low.x;
    range [3].y = high.y;
    range [3].z = (low.z + high.z) / 2;
}
///////////////////////////////////////////////////////////////////////
void 					MakeShapeBoundary
(
BentleyVectorDPoint3dR	range,
DPoint3dCR				low,
DPoint3dCR				high
)
{
	range.resize (5);
    range [0]	= low;
    range [4]	= low;

    range [2].x = high.x;
    range [2].y = high.y;
    range [2].z = high.z;

    range [1].x = high.x;
    range [1].y = low.y;
    range [1].z = (low.z + high.z) / 2;

    range [3].x = low.x;
    range [3].y = high.y;
    range [3].z = (low.z + high.z) / 2;
}
/////////////////////////////////////////////////////////////////////////////
///	<summary> 	OnUnload unload event handler.
///	<remarks>Unregister our dependency callback.</remarks></summary>
bool		OnUnload			(UnloadProgramReason   unloadType)
{
    DependencyManagerLinkage::RemoveRootsChangedCallback (DEPENDENCYAPPID_TestDepend);
	return false;
}
/////////////////////////////////////////////////////////////////////////////
///	<summary> 	OnReload reload event handler.</summary>
void		OnReload			(int argc, WCharCP argv[])
{
	WString	version (MicroStation::VersionString (TXT_DllName));
	WString announce;
	announce.Sprintf (L"Welcome to %s version %s", TXT_DllName, version.c_str ());
	const long	msgAttributes	{ 0 };
	NotifyMessageDetails	msg (OutputMessagePriority::Info, announce.c_str (),  announce.c_str (), msgAttributes, OutputMessageAlert::None);
	NotificationManager::OutputMessage (msg);
}
/** --------------------------------------------------------------------+
|																		|
|	function name:	cmdBoxDepend_create									|
|																		|
|	handles:		CMD_BOX_DEPEND_CREATE								|
|																		|
|	descr:			Creates a dependent box around selected elements	|
|																		|
|	returns:		Nothing												|
|																		|
+--------------------------------------------------------------------- */
void 				cmdBoxDepend_create
(
WCharCP				unused	//	=>	Possibly nullptr
)
{
	if (IsSelectionSetActive ())
	{
		WCharCP		msg	{ L"cmdBoxDepend_create" };
		mdlOutput_messageCenter (DgnPlatform::OutputMessagePriority::Debug, msg, msg, DgnPlatform::OutputMessageAlert::None);

		s_dependencyCallback.CreateBoxAround (SelectionSetManager::GetManager ());
	}
}
/** --------------------------------------------------------------------+
|																		|
|	function name:	cmdBoxDepend_drop									|
|																		|
|	handles:		CMD_BOX_DEPEND_DROP									|
|																		|
|	descr:			Detaches a dependent box from its root elements		|
|																		|
|	returns:		Nothing												|
|																		|
+--------------------------------------------------------------------- */
DLLEXPORT void 		cmdBoxDepend_drop
(
WCharCP				unused	//	=>	Possibly nullptr
)
{
	s_dependencyCallback.Drop ();
}
///////////////////////////////////////////////////////////////////////
///	<summary>cmdBoxDepend_offset.
///	<remarks>Implements command CMD_BOX_DEPEND_OFFSET_ID and CMD_BOX_DEPEND_OFFSET_ASSOC.
///	<para>Accept first member of selection set, then start
///	state machine to place cell at an XY offset from
///	its origin.
///	Store the offset in a DependencyLinkage.</para>
///	</remarks>
///	</summary>
DLLEXPORT  void 	cmdBoxDepend_offset
(
WCharCP				unused	//	=>	Possibly nullptr
)
{
	if (IsSelectionSetActive ())
	{
		DgnModelRefP	model	{ nullptr };
		ElementRefP		element	{ nullptr };
		if (SUCCESS == SelectionSetManager::GetManager ().GetElement (0, &element, &model))
		{
			WCharCP		ArbitraryCellName	{ L"CHAIR1" };
			ElementId	rootID	{ elementRef_getElemID (element) };
			//	What sort of dependency do we want to create?
			AppValues	dependencyType	= CMD_BOX_DEPEND_OFFSET_ID == mdlCommandNumber? TYPE_XYOFFSET_ID: TYPE_XYOFFSET_ASSOC;
			PlaceCellTool::InstallNewInstance (COMMAND_PlaceCellTool, PROMPT_PlaceCell, ArbitraryCellName, rootID, dependencyType);
		}
	}
}
/** --------------------------------------------------------------------+
|																		|
|	function name:	cmdBoxDepend_test									|
|																		|
|	handles:		CMD_BOX_DEPEND_TEST									|
|																		|
|	descr:			Hidden test function								|
|																		|
|	returns:		Nothing												|
|																		|
+--------------------------------------------------------------------- */
DLLEXPORT  void 	cmdBoxDepend_test
(
WCharCP				s	//	=>	Possibly nullptr
)
{
	if (s && wcslen (s))
	{
	}
}
/** --------------------------------------------------------------------+
|																		|
|	function name:	cmdBoxDepend_exit									|
|																		|
|	handles:		CMD_BOX_DEPEND_EXIT									|
|																		|
|	descr:			Unload this application								|
|																		|
|	returns:		Nothing												|
|																		|
+--------------------------------------------------------------------- */
DLLEXPORT  void 	cmdBoxDepend_exit
(
WCharCP				unused
)
{
	MicroStation::QueueMdlUnload ();
}
/*----------------------------------------------------------------------------------+
///	<summary>The main entry point for the native code MDL application.</summary>
///	<param name="argc">The count of args passed to the application.</param>
///	<param name="argv[]">The array of args passed to this application as wide char arrays.</param>
+-----------------------------------------------------------------------------------*/
extern "C" DLLEXPORT void MdlMain (int argc, WCharCP argv[])
{
    // Map key-in to function
    static MdlCommandNumber cmdNumbers[] =
    {
		{(CmdHandler)cmdBoxDepend_create,  	    CMD_BOX_DEPEND_CREATE 		},
		{(CmdHandler)cmdBoxDepend_drop,  	    CMD_BOX_DEPEND_DROP 		},
		{(CmdHandler)cmdBoxDepend_offset,  	    CMD_BOX_DEPEND_OFFSET_ID	},
		{(CmdHandler)cmdBoxDepend_offset,  	    CMD_BOX_DEPEND_OFFSET_ASSOC	},

        {(CmdHandler)cmdBoxDepend_test,  		CMD_BOX_DEPEND_TEST 		},
        {(CmdHandler)cmdBoxDepend_exit,  		CMD_BOX_DEPEND_EXIT 		},
        0,
    };

	MicroStation::MdlMain (cmdNumbers, MESSAGELIST_Commands, MESSAGELIST_Prompts);
	MicroStation::SetVersion (TXT_DllName);
	//  This call doesn't work!
	VersionNumber   	msVersion	{ 1, 2, 3, 4};
	mdlSystem_setMdlAppVersionNumber (nullptr, &msVersion);

	SystemCallback::SetUnloadProgramFunction (OnUnload);
	SystemCallback::SetReloadProgramFunction (OnReload);
	OnReload (argc, argv);
	//	Our dependency registration
    DependencyManagerLinkage::RegisterRootsChangedCallback (DEPENDENCYAPPID_TestDepend,
    	&s_dependencyCallback);
}
