#include	"BoxDependency.h"
#include	"BoxDependencyIds.h"
#include	"BoxDependencyCallback.h"
#include 	"MessageTracer.h"
#include	<MicroStationUtilities.h>
#include	<Mstn/ISessionMgr.h>
#include    <Mstn/MdlApi/MdlApi.h>
#include 	<DgnPlatform/LinearHandlers.h>
#include 	<DgnPlatform/DropGraphics.h>
#include 	<DgnPlatform/AssociativePoint.h>

USING_NAMESPACE_BENTLEY;
USING_NAMESPACE_BENTLEY_DGNPLATFORM;
USING_NAMESPACE_BENTLEY_MSTNPLATFORM;
USING_NAMESPACE_BENTLEY_MSTNPLATFORM_ELEMENT;

///////////////////////////////////////////////////////////////////////
int							BoxDependencyCallback::OnSelfStatusChange
(
int							selfStatus,
DPoint3dCR					origin,		//	=>	Of root element
DVec3dCR					storedOffset,
ElementHandleCR				dependentElement,
IDependencyOffsetLinkageP	linkage		//	=>	OffsetAssocDependencyLinkage or OffsetIdDependencyLinkage
)
{
	CMessageTracer			msg (L"OnSelfStatusChange rootStatus="); msg.Append (selfStatus);
	msg.Output ();
	int						rc		{ SUCCESS };
	DPoint3d				placed;
	DisplayHandlerP	handler = dependentElement.GetDisplayHandler ();
	handler->GetSnapOrigin (dependentElement, placed);

	switch (selfStatus)
	{
		case DEPENDENCY_STATUS_DELETED:
			return SUCCESS;
			break;
		case DEPENDENCY_STATUS_CHANGED:
		{
			//	Dependent element may have moved: re-compute offset and update linkage
			DVec3d 					currentOffset;
			currentOffset.DifferenceOf (placed, origin);
			//	Pointer to offset placed earlier in existing linkage
			///	<remarks>
			///	Check to see whether the current offset is
			///	different to the offset stored in the linkage.
			///	If we don't perform this test before updating the
			///	linkage, we get called repeatedly even when
			///	there's no change.  Replacing the linkage
			///	modifies this dependent element, triggering the
			///	dependency API to invoke this callback again
			///	</remarks>
			const double& Tolerance	= ACTIVEMODEL->GetModelInfoCP ()->GetUorPerMeter ();
			//	AlmostEqual has too tight a tolerance
			//if (!currentOffset.AlmostEqual (storedOffset))
			if (!currentOffset.IsEqual (storedOffset, Tolerance))
			{
				msg = L"storedOffset "; msg.Append (storedOffset);
				msg.Append (" currentOffset "); msg.Append (currentOffset);
				msg.Append (currentOffset.IsEqual (storedOffset, Tolerance)? L" zero offset": L" significant offset");
				msg.Warning (DgnPlatform::OutputMessagePriority::Warning, false);
				linkage->SetOffset (currentOffset);
				linkage->ReplaceInModel (dependentElement);
			}
			break;
		}
		case DEPENDENCY_STATUS_UNCHANGED:
			break;
	}
	return rc;
}
///////////////////////////////////////////////////////////////////////
int							BoxDependencyCallback::OnRootStatusChange
(
int							rootStatus,		//	=> pRootStatus[0]
DPoint3dCR					origin,			//	=>	Of root element
DVec3dCR					offset,
ElementHandleCR				dependentElement
)
{
	int						rc		{ SUCCESS };
	CMessageTracer			msg (L"OnRootStatusChange rootStatus="); msg.Append (rootStatus);
	msg.Output ();

	DPoint3d				placed;
	DisplayHandlerP	handler = dependentElement.GetDisplayHandler ();
	handler->GetSnapOrigin (dependentElement, placed);

    switch (rootStatus)
    {
	case DEPENDENCY_STATUS_DELETED:
	{	//	Delete dependent element
		EditElementHandle	deleteCandidate (dependentElement.GetElementId (), ACTIVEMODEL);
		deleteCandidate.DeleteFromModel ();
		break;
	}
	case DEPENDENCY_STATUS_CHANGED:
	{
		msg = L"DEPENDENCY_STATUS_CHANGED: root element moved";
		msg.Output ();
		//	Move dependent element relative to root element
		//msg = L"origin of root element ";	msg.Append (rootOrigin);
		//msg += L" current location of dependent element ";	msg.Append (placed);
		//msg += L" stored offset ";	msg.Append (storedOffset);
		//msg.Output ();
		//	Translate dependent element using offset
    	DVec3d				local;
    	local.DifferenceOf (origin, placed);
    	local.Add (offset);
		Transform			transform;
		transform.InitIdentity ();
		transform.InitFrom (local);
		transform.SetTranslation (local);
		TransformInfo 		transformInfo (transform);
    	const bool&			DuplicateDescriptor	{ true };
    	EditElementHandle	replacement (dependentElement, DuplicateDescriptor);
    	Handler&			handler	= replacement.GetHandler (MISSING_HANDLER_PERMISSION_Transform);
		handler.ApplyTransform (replacement, transformInfo);
		rc	= replacement.ReplaceInModel (dependentElement.GetElementRef ());
		if (SUCCESS == rc)
		{
		}
		else
		{
			msg = L"replacement.ReplaceInModel failed rc="; msg.Append (rc);
			msg.Warning (DgnPlatform::OutputMessagePriority::Warning, false);
		}
		break;
	}
	case DEPENDENCY_STATUS_UNCHANGED:
		//copy the elemid from the existing to the new.
		//newLinkage.AddRootId (linkage.root.elemid[0]);
		//	and now what?
		break;
	}
	return rc;
}
///////////////////////////////////////////////////////////////////////
int							BoxDependencyCallback::ProcessOffsetAssocCallback
(
ElementHandleCR				dependentElement,
DependencyLinkageCR			linkage,
UChar*						pRootStatus,
UChar		    			selfStatus
)
{
	OffsetAssocDependencyLinkage	newLinkage 		(linkage);
	CMessageTracer			msg (L"ProcessOffset Assoc Callback ");
	msg.Output ();
    int						rc				{ SUCCESS };
	DPoint3d				placed;
	DisplayHandlerP	handler = dependentElement.GetDisplayHandler ();
	handler->GetSnapOrigin (dependentElement, placed);
	//	Get origin from association point
	AssocPointCR			assocPoint	= newLinkage.NthAssocPoint (0);
	DPoint3d				origin;
	AssociativePoint::GetPoint (&origin, assocPoint, ACTIVEMODEL);
	msg	= L"Stored offset "; msg.Append (newLinkage.GetOffset ());
	msg	+= L"Assoc point origin  ";  msg.Append (origin);
	msg.Output ();
	OnSelfStatusChange (selfStatus, origin, newLinkage.GetOffset (), dependentElement, &newLinkage);
	rc	= OnRootStatusChange (pRootStatus[0], origin, newLinkage.GetOffset (), dependentElement);
	return rc;
}
///////////////////////////////////////////////////////////////////////
int							BoxDependencyCallback::ProcessOffsetIdCallback
(
ElementHandleCR				dependentElement,
DependencyLinkageCR			linkage,
UChar*						pRootStatus,
UChar		    			selfStatus
)
{
	OffsetIdDependencyLinkage	newLinkage 		(linkage);
	CMessageTracer			msg (L"ProcessOffset ID Callback root ID="); msg.ElementId (linkage.root.elemid[0]);
	//msg += L" newLinkage root ID="; msg.ElementId (newLinkage.NthRootId (0));
	//msg += L" Self Status="; msg.Append (selfStatus);
	//msg += L" Extra Bytes="; msg.Append ((int)newLinkage.ExtraBytes ());
	//msg.Output ();
    int						rc				{ SUCCESS };
	DPoint3d				placed;
	DisplayHandlerP	handler = dependentElement.GetDisplayHandler ();
	handler->GetSnapOrigin (dependentElement, placed);
	//	Get origin of root element
	DPoint3d				rootOrigin;
	GetElementOriginById (rootOrigin, newLinkage.NthRootId (0));
	//msg	= L"Stored offset "; msg.Append (newLinkage.GetOffset ());
	//msg.Output ();

	OnSelfStatusChange (selfStatus, rootOrigin, newLinkage.GetOffset (), dependentElement, &newLinkage);
	rc	= OnRootStatusChange (pRootStatus[0], rootOrigin, newLinkage.GetOffset (), dependentElement);

    /*-------------------------------------------------------------------
	Root deleted?
    -------------------------------------------------------------------*/
    if  (newLinkage.NumRoots () != linkage.nRoots)
    {
		newLinkage.ReplaceInModel (dependentElement);
		//ReplaceOffsetDependencyLinkage (dependentElement, filePos, &newLinkage);
	}

    return  rc;
}
///////////////////////////////////////////////////////////////////////
int							BoxDependencyCallback::ProcessBoxCallback
(
ElementHandleCR	    		dependentElement,
DependencyLinkageCR			linkage,
UChar*						pRootStatus,
UChar		    			selfStatus
)
{
	BoxDependencyLinkage	boxLinkage (linkage);
	CMessageTracer			msg (L"BoxDependencyCallback::ProcessBoxCallback nLinks="); msg.Append (boxLinkage.NumRoots ());
	msg	+= L" App ID "; msg.Append (boxLinkage.GetAppId ()); msg += L" App Value "; msg.Append (boxLinkage.GetAppValue ());
	msg.Output ();
	switch (selfStatus)
	{
		case DEPENDENCY_STATUS_DELETED:
			return SUCCESS;
			break;
		case DEPENDENCY_STATUS_CHANGED:
			//	Allow box to be moved or resized manually
			return SUCCESS;
			//	If we don't return here, the box is re-created
			//	each time it's modified manually
			break;
		case DEPENDENCY_STATUS_UNCHANGED:
			break;
	}

    for (UInt16 i = 0; i != linkage.nRoots; ++i)
    {
    	switch (pRootStatus[i])
    	{
		case DEPENDENCY_STATUS_DELETED:
			break;
		case DEPENDENCY_STATUS_CHANGED:
		case DEPENDENCY_STATUS_UNCHANGED:
	    	//copy the elemid from the root elements to the new.
	    	boxLinkage.AddRootId (linkage.root.elemid[i]);
	    	break;
	    }
	}
	msg = L"Added root IDs nLinks="; msg.Append (boxLinkage.NumRoots ());
	msg.Output ();
    /*-------------------------------------------------------------------
	Root deleted?
    -------------------------------------------------------------------*/
    if (boxLinkage.NumRoots () != linkage.nRoots)
    {
		EditElementHandle	eeh (dependentElement.GetElementId (), ACTIVEMODEL);
		//boxLinkage.DeleteLinkage (eeh);
		boxLinkage.AppendLinkage (eeh);
		eeh.ReplaceInModel (dependentElement.GetElementRef ());
	}

    /*-------------------------------------------------------------------
	Root changed or box modified
    -------------------------------------------------------------------*/
    switch  (linkage.appValue)
    {
		case TYPE_BOX:
		{
	    	//const DependencyLinkage *pData	{ boxLinkage.GetLinkage (dependentElement) };
	    	//DependencyManagerLinkage::GetLinkage (&pData, dependentElement, linkage.appID, linkage.appValue);
	    	RecomputeBox (&dependentElement, boxLinkage);
	    	break;
		}
	}

    return  SUCCESS;
}
///////////////////////////////////////////////////////////////////////
void					BoxDependencyCallback::RecomputeBox
(
ElementHandleCP			eh,
BoxDependencyLinkageCR	linkage
)
{
    CMessageTracer	msg	(L"BoxDependencyCallback::RecomputeBox linkage.NumRoots="); msg.Append (linkage.NumRoots ());
    msg.Output ();
    DPoint3d    low			{ DBL_MAX, DBL_MAX, DBL_MAX };
    DPoint3d	high		{ -(DBL_MAX-1), -(DBL_MAX-1), -(DBL_MAX-1) };

    for (UInt16 iRoot = 0; iRoot != linkage.NumRoots (); ++iRoot)
    {
		ElementHandle root (linkage.NthRootId (iRoot), ACTIVEMODEL);
		if (MicroStation::IsGraphicalElement (root))
		{
    		DRange3d   	range;
			mdlElement_extractRange (&range, root.GetElementCP ());
	    	FindHighLow (low, high, range);
	    }
	}

    const int	nCorners	{ 5 };
    bvector<DPoint3d>	corners (nCorners);
    MakeShapeBoundary (corners, low, high);
  	const bool& 						ForceXYOrientation	{ false };
  	const CurveVector::BoundaryType&	BoundaryType		{ CurveVector::BOUNDARY_TYPE_Outer  };
	CurveVectorPtr		shapeVector = CurveVector::CreateLinear (corners, BoundaryType, ForceXYOrientation);
    EditElementHandle	eeh;
	DraftingElementSchema::ToElement (eeh, *shapeVector, eh, ACTIVEMODEL->Is3d (), *ACTIVEMODEL);
	if (nullptr == eh)
	{
		msg	= L"No existing element: create new box and append linkage";
		msg.Output ();
		ElementPropertiesSetterPtr props	= ElementPropertiesSetter::Create ();
		props->SetColor (3);
		props->Apply (eeh);
		linkage.AppendLinkage (eeh);
		eeh.AddToModel ();
	}
	else
	{
		msg	= L"Update existing element";
		msg.Output ();
		linkage.AppendLinkage (eeh);
		eeh.ReplaceInModel (eh->GetElementRef ());
	}
	nBoxElementId_	= eeh.GetElementId ();
}
///////////////////////////////////////////////////////////////////////
void					BoxDependencyCallback::CreateBoxAround
(
SelectionSetManagerR	selection
)
{
	//	Create a dependency linkage that points to all elements in the selection set
    BoxDependencyLinkage	boxLinkage;
    for (size_t i = 0; i != selection.NumSelected (); ++i)
    {
		ElementRefP		member	{ nullptr	};
		DgnModelRefP	model	{ nullptr	};
		selection.GetElement (i, &member, &model);
		if (mdlModelRef_isActiveModel (model))
		{
			boxLinkage.AddRootId (elementRef_getElemID (member));
		}
	}

    RecomputeBox (nullptr, boxLinkage);
}
///////////////////////////////////////////////////////////////////////
void	BoxDependencyCallback::Drop	()
{
    EditElementHandle	eeh	(nBoxElementId_, ACTIVEMODEL);
	enum LinkageInfo {	ExpectedNumberOfLinkages	= 1, };
    if (ExpectedNumberOfLinkages != DependencyManagerLinkage::DeleteLinkage (eeh, DEPENDENCYAPPID_TestDepend, TYPE_BOX))
    {
		WCharCP		msg	{ L"Linkage not found!" };
		mdlOutput_messageCenter (DgnPlatform::OutputMessagePriority::Warning, msg, msg, DgnPlatform::OutputMessageAlert::Balloon);
	}

    eeh.ReplaceInModel (eeh.GetElementRef ());
    nBoxElementId_ = 0;
}
///////////////////////////////////////////////////////////////////////
StatusInt  BoxDependencyCallback::OnRootsChanged	(ElementHandleCR			dependentElement,
													DependencyLinkage const&	dependencyData,
													UInt8*						pRootStatus,
													UInt8 						selfStatus)
{
	CMessageTracer	msg ("OnRootsChanged appValue="); msg.Append (dependencyData.appValue);
	msg	+= L" App ID "; msg.Append (dependencyData.appID);
	msg.Output ();
    switch  (dependencyData.appValue)
    {
		case TYPE_BOX:
			return ProcessBoxCallback (dependentElement, dependencyData, pRootStatus, selfStatus);
			break;
		case TYPE_XYOFFSET_ID:
			return ProcessOffsetIdCallback (dependentElement, dependencyData, pRootStatus, selfStatus);
			break;
		case TYPE_XYOFFSET_ASSOC:
			return ProcessOffsetAssocCallback (dependentElement, dependencyData, pRootStatus, selfStatus);
			break;
	}
	return SUCCESS;
}
///////////////////////////////////////////////////////////////////////
WString		BoxDependencyCallback::GetDescription () const
{
	return WString (L"BoxDependency callback");
}
///////////////////////////////////////////////////////////////////////
