/*-------------------------------------------------------------------- +
|
|   BoxDependencyCallback.h
|
+--------------------------------------------------------------------- */
/*-------------------------------------------------------------------- +
|																		|
| Copyright (C) 2017 LA Solutions Ltd, 									|
| All rights reserved.													|
|																		|
+----------------------------------------------------------------------*/
#if !defined(BOX_DEPENDENCY_CALLBACK_H_INCLUDED_)
#define BOX_DEPENDENCY_CALLBACK_H_INCLUDED_
#pragma once

#include    <StaticIRootsChangedCallback.h>
#include	"BoxDependencyLinkage.h"

///////////////////////////////////////////////////////////////////////
///	<summary>BoxDependencyCallback class.
///	<remarks>See Be Communities post
///	http://communities.bentley.com/communities/other_communities/bdn_other_communities/b/bdn-blog/archive/2016/12/27/ce-c-element-dependency.
/// </remarks>
///	</summary>
struct BoxDependencyCallback : StaticIRootsChangedCallback
{
	///	<summary>OnRootStatusChange called from OnRootsChanged to handle a root event.</summary>
	///	<param name="rootStatus" >Since our Offset examples have only a single root, its status (pRootStatus[0]) is passed here.</param>
	///	<param name="origin" >The root origin.  May be calculated or obtained from an association point.</param>
	///	<param name="offset" >The offset vector of the dependent element from its root.</param>
	///	<param name="dependentElement" >The read-only dependent element.</param>
	int							OnRootStatusChange
	(
	int							rootStatus,
	DPoint3dCR					origin,
	DVec3dCR					offset,
	ElementHandleCR				dependentElement
	);
	///	<summary>OnSelfStatusChange called from OnRootsChanged to handle a dependent element event.</summary>
	///	<param name="selfStatus" >The status of the dependent element.</param>
	///	<param name="origin" >The root origin.  May be calculated or obtained from an association point.</param>
	///	<param name="offset" >The offset vector of the dependent element from its root.</param>
	///	<param name="dependentElement" >The read-only dependent element.</param>
	///	<param name="linkage" >A pointer to our class that inherits from IDependencyOffsetLinkage.</param>
	int							OnSelfStatusChange
	(
	int							selfStatus,
	DPoint3dCR					origin,
	DVec3dCR					offset,
	ElementHandleCR				dependentElement,
	IDependencyOffsetLinkageP	linkage		//	=>	OffsetAssocDependencyLinkage or OffsetIdDependencyLinkage
	);
	///	<summary>The ID of the box we create around selected elements.</summary>
	Bentley::DgnPlatform::ElementId		nBoxElementId_;

	///	<summary>	OnRootsChanged() does the work when MicroStation observes a dependency change.
	///	<remarks>This callback is notified to MicroStation using DependencyManagerLinkage::RegisterRootsChangedCallback (),
	///	probably in MdlMain().
	///	<para>MicroStation calls us back when it observes a dependency event.  The event
	///	types is notified in the rootStatus and selfStatus parameters.</para></remarks>
	///	<param name="dependentElement" >The DGN element that is involved.</param>
	///	<param name="dependencyData" >The DependencyLinkage extracted from the element.</param>
	///	<param name="pRootStatus" >An array of rootStatus values, one for each dependency root.</param>
	///	<param name="selfStatus" >The status of the dependent element.</param>
	///	</summary>
	virtual StatusInt  OnRootsChanged				(ElementHandleCR								dependentElement,
													Bentley::DgnPlatform::DependencyLinkage const&	dependencyData,
													UInt8*											pRootStatus,
													UInt8 											selfStatus) override;
	///	<summary>	This description is used for debugging.</summary>
	virtual Bentley::WString  						GetDescription () const override;
	///	<summary>	ProcessOffsetIdCallback called by OnRootsChanged.</summary>
	///	<param name="dependentID">Element ID of dependent element.</param>
	///	<param name="modelRef">Model of dependent element.</param>
	///	<param name="linkage">Copy of dependent element's linkage.</param>
	///	<param name="pRootStatus">Change status of each root in linkage: DEPENDENCY_STATUS_XXX.</param>
	///	<param name="selfStatus"></param>
	int												ProcessOffsetIdCallback
	(
	ElementHandleCR									dependentElement,
	DependencyLinkageCR								linkage,
	UChar*											pRootStatus,
	UChar		    								selfStatus
	);
	///	<summary>	ProcessOffsetAssocCallback called by OnRootsChanged.</summary>
	///	<param name="dependentID">Element ID of dependent element.</param>
	///	<param name="modelRef">Model of dependent element.</param>
	///	<param name="linkage">Copy of dependent element's linkage.</param>
	///	<param name="pRootStatus">Change status of each root in linkage: DEPENDENCY_STATUS_XXX.</param>
	///	<param name="selfStatus"></param>
	int												ProcessOffsetAssocCallback
	(
	ElementHandleCR									dependentElement,
	DependencyLinkageCR								linkage,
	UChar*											pRootStatus,
	UChar		    								selfStatus
	);
	///	<summary>	ProcessBoxCallback called by OnRootsChanged.</summary>
	///	<param name="dependentID">Element ID of dependent element.</param>
	///	<param name="modelRef">Model of dependent element.</param>
	///	<param name="linkage">Copy of dependent element's linkage.</param>
	///	<param name="pRootStatus">Change status of each root in linkage: DEPENDENCY_STATUS_XXX.</param>
	///	<param name="selfStatus"></param>
	int												ProcessBoxCallback
	(
	ElementHandleCR									dependentElement,
	DependencyLinkageCR								linkage,
	UChar*											pRootStatus,
	UChar		    								selfStatus
	);
	///	<summary>	RecomputeBox from extent of root elements in linkage.</summary>
	///	<param name="eh">Dependent element.  May be nullptr. </param>
	///	<param name="linkage">Reference to dependent element's linkage.</param>
	void											RecomputeBox
	(
	ElementHandleCP									dependentElement,
	BoxDependencyLinkageCR							linkage
	);
	///	<summary>	Compute a rectangle from extent of root elements in selecction set.</summary>
	void											CreateBoxAround
	(
	SelectionSetManagerR		selection
	);
	///	<summary>	Remove the box defined by our dependency.</summary>
	void											Drop ();

	BoxDependencyCallback	()	: nBoxElementId_ (0) {}
};


#endif	// !defined(BOX_DEPENDENCY_CALLBACK_H_INCLUDED_)
