/*-------------------------------------------------------------------- +
|
|   BoxDependencyLinkage.h
|
+--------------------------------------------------------------------- */
/*-------------------------------------------------------------------- +
|																		|
| Copyright (C) 2017 LA Solutions Ltd, 									|
| All rights reserved.													|
|																		|
+----------------------------------------------------------------------*/
#if !defined(BOX_DEPENDENCY_LINKAGE_H_INCLUDED_)
#define BOX_DEPENDENCY_LINKAGE_H_INCLUDED_
#pragma once

#include 	<type_traits>
#include 	<DgnPlatform/DependencyManagerLinkage.h>
#include	<Mstn/MdlApi/elementref.h>
#include	"BoxDependencyIds.h"

///////////////////////////////////////////////////////////////////////
///	<summary>Base class for our DependencyLinkages.</summary>
///	<remarks>This class specifies the DEPENDENCYAPPID_TestDepend, and is used
///	for Element ID dependency types.  It is pre-configured to drop the dependency
///	if an element is copied.
///	<para>DEPENDENCYAPPID_TestDepend is a macro in header DependencyManagerAppId.h.</para>
///	</remarks>
///	<param name="AppId" >Developer ID assigned by Bentley Systems.  In this example, DEPENDENCYAPPID_TestDepend.</param>
///	<param name="AppType" >Application-specific ID.  In this example, we define enum AppValues (TYPE_BOX, TYPE_XYOFFSET) in BoxDependencyIds.h.</param>
///	<param name="DataType" >DEPENDENCY_DATA_TYPE_XXX.  The type of data to be stored in the dependency linkage: Element IDs, association points etc.</param>
///	<param name="CopyPolicy" >DEPENDENCY_ON_COPY_XXX.  What happens when user copies a dependent element.</param>
template <class T, int AppId, int AppType, int DataType, int CopyPolicy> class DependencyLinkageBase
	:  public Bentley::DgnPlatform::DependencyLinkage //, public IDependencyOffsetLinkage
{
protected:
	///	<summary>	Extra bytes of data to be appended after the last root data.
	///	<remarks>	These data are application-dependent.  You may not want any extra bytes.</remarks></summary>
	const size_t  	xtrabytes_;

public:
	DependencyLinkageBase (size_t xtrabytes = 0) : xtrabytes_ (xtrabytes)
	{
   		Bentley::DgnPlatform::DependencyManagerLinkage::InitLinkage (*this, AppId, DataType, CopyPolicy);
		nRoots		= 0;
   		appValue	= static_cast<UShort>(AppType);
   	}
   	///	<summary>Copy constructor accepts a MicroStationAPI DependencyLinkage.</summary>
	explicit DependencyLinkageBase (Bentley::DgnPlatform::DependencyLinkage const& linkage, size_t extraBytes = 0)
	:	Bentley::DgnPlatform::DependencyLinkage (linkage),
		xtrabytes_	(extraBytes)
	{
   	}
   	///	<summary>Assignment operator accepts a MicroStationAPI DependencyLinkage.</summary>
   	DependencyLinkageBase&	operator=(Bentley::DgnPlatform::DependencyLinkage const& linkage)
   	{
		DependencyLinkageBase	temp (linkage);
		std::swap (*this, temp);
		return *this;
	}
	///	<summary>Get the number of root elements in this dependency.</summary>
	UInt16	NumRoots		()				const	{	return nRoots; }

	///	<summary>Helper function deletes this linkage from an element.</summary>
	bool	DeleteLinkage	(Bentley::EditElementHandleR	eeh)	const
	{
		return SUCCESS == DependencyManagerLinkage::DeleteLinkage (eeh, appID, appValue);
	}
	///	<summary>Helper function adds this linkage to an element.</summary>
	bool	AppendLinkage	(Bentley::EditElementHandleR	eeh)	const
	{
		return SUCCESS == DependencyManagerLinkage::AppendLinkage (eeh, *this, xtrabytes_);
	}
	///	<summary>Helper function updates this linkage in an element.</summary>
	bool	UpdateLinkage	(Bentley::EditElementHandleR	eeh)	const
	{
		return SUCCESS == DependencyManagerLinkage::UpdateLinkage (eeh, *this, xtrabytes_);
	}
	///	<summary>ReplaceInModelImpl updates the linkage on an existing element and replaces that element.</summary>
	void	ReplaceInModelImpl (ElementHandleCR	eh) const
	{
		const bool&				DuplicateDescriptor	{ true };
		EditElementHandle		replacement (eh, DuplicateDescriptor);
		UpdateLinkage (replacement);
		replacement.ReplaceInModel (eh.GetElementRef ());
	}
	///	<summary>Helper function gets a pointer to a linkage on an element.
	///	<remarks>static.<para>Pointer returned value could be nullptr!</para></remarks></summary>
	DependencyLinkage const*	GetLinkage (Bentley::ElementHandleCR	eh) const
	{
		const DependencyLinkage *pData	{ nullptr };
		DependencyManagerLinkage::GetLinkage (&pData, eh, appID, appValue);
		return pData;
	}
	///	<summary>Return the App ID of this linkage.</summary>
	UInt16 const&	GetAppId 		()	const	{	return appID;		}
	///	<summary>Return the App Value of this linkage.</summary>
	UInt16 const&	GetAppValue		()	const	{	return appValue;	}
	///	<summary>Reveal the amount of extra storage.</summary>
	size_t			GetExtraBytes	()	const	{	return xtrabytes_;	}


	///////////////////////////////////////////////////////////////////
	///	<summary>C++ meta programming constant type definitions.</summary>
	///	<remarks>Other DEPENDENCY_DATA_TYPE_XXX not defined since we're not using them.</remarks>
	using t_id		= std::integral_constant <UInt16, DEPENDENCY_DATA_TYPE_ELEM_ID>;
	using t_assoc	= std::integral_constant <UInt16, DEPENDENCY_DATA_TYPE_ASSOC_POINT>;
	using t_data	= std::integral_constant <UInt16, DataType>;
	///////////////////////////////////////////////////////////////////
	///	<summary>C++ meta programming conditional tests.</summary>
	///	<remarks>We're not using these -- just testing whether the compiler accepts them.</remarks>
	using x1 = std::is_same <t_id, t_data>;
	using x2 = std::is_same <t_assoc, t_data>;
	using enable1 = typename std::enable_if<x1::value>;
	using enable2 = typename std::enable_if<x2::value>;
	///	<summary>C++ meta programming compiles function depending on DataType.</summary>
	//int trivial (std::enable_if <x1::value,bool>) { return 1; }	//compiles OK but requires argument when instantiated
	///	<summary>Template meta-functions are enabled/disabled depending on DEPENDENCY_DATA_TYPE_XXX.</summary>
	int trivial (enable1 = 0) { return 1; }
	int trivial (enable2 = 0) { return 2; }
	//	From StackOverflow
	//	http://stackoverflow.com/questions/14600201/why-should-i-avoid-stdenable-if-in-function-signatures
	//	https://en.wikibooks.org/wiki/More_C%2B%2B_Idioms/enable-if
	///	<summary> read2 OK but requires explicit template instantiation: read2<DEPENDENCY_DATA_TYPE_ELEM_ID>()</summary>
	template<int DataType2>
	int read2(typename std::enable_if<
		  std::is_same<t_id, std::integral_constant <UInt16, DataType2>>::value >::type* = 0) { return DataType2; }
	///	<summary> read2 OK but requires explicit template instantiation: read2<DEPENDENCY_DATA_TYPE_ASSOC_POINT>()</summary>
	template<int DataType2>
	int read2(typename std::enable_if<
		  std::is_same<t_assoc, std::integral_constant <UInt16, DataType2>>::value >::type* = 0) { return DataType2; }
	///	<summary> read3 OK and defaults to the DataType supplied to this class template.</summary>
	///	<remarks>This specialisation matches t_id defined above when DataType == DEPENDENCY_DATA_TYPE_ELEM_ID.</remarks>
	template<int DataType3 = DataType>
	int read3(typename std::enable_if<
		  std::is_same<t_id, std::integral_constant <UInt16, DataType3>>::value >::type* = 0) { return DataType3; }
	///	<summary> read3 OK and defaults to the DataType supplied to this class template.</summary>
	///	<remarks>This specialisation matches t_assoc defined above when DataType == DEPENDENCY_DATA_TYPE_ASSOC_POINT.</remarks>
	template<int DataType3 = DataType>
	int read3(typename std::enable_if<
          std::is_same<t_assoc, std::integral_constant <UInt16, DataType3>>::value >::type* = 0) { return DataType3; }
	///////////////////////////////////////////////////////////////////
	///	<summary>The following specialisations are enabled when our DataType is DEPENDENCY_DATA_TYPE_ELEM_ID.</summary>
	///	<summary>Get the Element ID of the Nth root element in this dependency.</summary>
	///	<remarks>Method is valid only if DEPENDENCY_DATA_TYPE_ELEM_ID.</remarks>
	template<int DataType3 = DataType>
	Bentley::DgnPlatform::ElementId const&		NthRootId		(UInt16	index,
		typename std::enable_if<
          std::is_same<t_id, std::integral_constant <UInt16, DataType3>>::value >::type* = 0)	const
    {	return root.elemid[index]; }
	///	<summary>Convenience function to get the first root ID.</summary>
	template<int DataType3 = DataType>
	Bentley::DgnPlatform::ElementId const&		FirstRootId		(typename std::enable_if<
          std::is_same<t_id, std::integral_constant <UInt16, DataType3>>::value >::type* = 0)	const
  	{	return root.elemid[0]; }
	///	<summary>Add a root element ID to a dependent element's linkage.
	///	<remarks>In this example we assume that all elements are in the active model.
	///	<para>Method is valid only if DEPENDENCY_DATA_TYPE_ELEM_ID.</para>
	///	</remarks></summary>
	template<int DataType3 = DataType>
   	void	AddRootId	(Bentley::ElementRefP elemRef,
   		typename std::enable_if<
          std::is_same<t_id, std::integral_constant <UInt16, DataType3>>::value >::type* = 0)
   	{
		AddRootId (elementRef_getElemID (elemRef));
	}
	template<int DataType3 = DataType>
   	void	AddRootId	(Bentley::DgnPlatform::ElementId const&	id,
   		typename std::enable_if<
          std::is_same<t_id, std::integral_constant <UInt16, DataType3>>::value >::type* = 0)
   	{
		root.elemid[nRoots++] = id;
	}
	///////////////////////////////////////////////////////////////////
	///	<summary>The following specialisations are enabled when our DataType is DEPENDENCY_DATA_TYPE_ASSOC_POINT.</summary>
	///	<summary>Add a root association to a dependent element's linkage.
	///	<remarks>In this example we assume that all elements are in the active model.
	///	<para>Method is valid only if DEPENDENCY_DATA_TYPE_ASSOC_POINT.</para>
	///	</remarks></summary>
	template<int DataType3 = DataType>
	void	AddRootAssoc	(AssocPointCR	assocPoint,
   		typename std::enable_if<
          std::is_same<t_assoc, std::integral_constant <UInt16, DataType3>>::value >::type* = 0)
	{
		root.assoc[nRoots++] = assocPoint;
	}
	///	<summary>Get the AssocPoint of the Nth root element in this dependency.</summary>
	///	<remarks>Method is valid only if DEPENDENCY_DATA_TYPE_ASSOC_POINT.</remarks>
	template<int DataType3 = DataType>
	AssocPointCR		NthAssocPoint	(UInt16	index,
   		typename std::enable_if<
          std::is_same<t_assoc, std::integral_constant <UInt16, DataType3>>::value >::type* = 0)	const
    {	return root.assoc[index]; }
	///	<summary>Convenience function to get the first assoc point.</summary>
	template<int DataType3 = DataType>
	AssocPointCR		FirstAssocPoint	(
   		typename std::enable_if<
          std::is_same<t_assoc, std::integral_constant <UInt16, DataType3>>::value >::type* = 0)	const
   	{	return root.assoc[0]; }

};	//	DependencyLinkageBase
///////////////////////////////////////////////////////////////////////
///	<summary>IDependencyOffsetLinkage.</summary>
///	<remarks>Inheritable interface lets us pass an IDendencyOffset*.</remarks.
struct IDependencyOffsetLinkage
{
	///	<returns>Get the offset vector.</returns>
	virtual DVec3d const& 	GetOffset		()	const = 0;
	///	<summary>SetOffset.</summary>
	///	<returns>Set the offset vector.</returns>
	virtual void			SetOffset		(DVec3d const&					offset) = 0;
	///	<summary>UpdateLinkage on an element.</summary>
	//virtual bool			UpdateLinkage	(Bentley::EditElementHandleR	eeh)	const = 0;
	///	<summary>ReplaceInModel updates the linkage on an element and rewrites that element.</summary>
	virtual void			ReplaceInModel (ElementHandleCR	eh) const = 0;
	/*
	virtual void			ReplaceInModel (ElementHandleCR	eh)
	{
		const bool&				DuplicateDescriptor	{ true };
		Bentley::DgnPlatform::EditElementHandle		replacement (eh, DuplicateDescriptor);
		UpdateLinkage (replacement);
		replacement.ReplaceInModel (eh.GetElementRef ());
	}
	*/
};

using IDependencyOffsetLinkageP		= IDependencyOffsetLinkage*;
using IDependencyOffsetLinkageCP	= IDependencyOffsetLinkage const *;
///////////////////////////////////////////////////////////////////////
///	<summary>Helper classes for our DependencyLinkages.
///	<remarks>OffsetDependencyLinkage  is a specialisation of DependencyLinkageBase for TYPE_XYOFFSET.
///	<para>It has additional functionality to store the DVec3d offset vector.</para>
///	<para>DVec3d occupies unused root ID storage space, since this class needs only one root ID.</para>
///	</remarks></summary>
struct OffsetIdDependencyLinkage :  public DependencyLinkageBase<OffsetIdDependencyLinkage, DEPENDENCYAPPID_TestDepend, TYPE_XYOFFSET_ID, DEPENDENCY_DATA_TYPE_ELEM_ID, DEPENDENCY_ON_COPY_DropDependency>
									, public IDependencyOffsetLinkage
{
	///	<summary>GetOffset pointer from a dependency linkage.
	///	<remarks>static</remarks></summary>
	static DVec3d* GetOffset (DependencyLinkageCR			linkage)
	{
		return (DVec3d*)&linkage.root.elemid[1];
	}
	///	<summary>GetOffset from this object.</summary>
	///	<returns>Get the offset vector.</returns>
	virtual DVec3d const& 	IDependencyOffsetLinkage::GetOffset	()	const override
	{
		return *GetOffset (*this);
	}
	///	<summary>ReplaceInModel updates an existing element with the current state of this linkage.</summary>
	///	<remarks>Forwards to base class implementation.</remarks>
	void			ReplaceInModel (ElementHandleCR	eh) const override
	{
		ReplaceInModelImpl (eh);
	}
	///	<summary>SetOffset.</summary>
	///	<returns>Set the offset vector.</returns>
	virtual void		IDependencyOffsetLinkage::SetOffset		(DVec3d const&	offset) override
	{
		//	Nasty C++ hack
		DVec3d* target	= reinterpret_cast<DVec3d*>(&root.elemid[1]);
		*target = offset;
		//	But MDL requires an equally-nasty C-style memcpy
		//memcpy (&root.elemid[1], &offset,  xtrabytes_);
	}
	///	<summary>Copy constructor taking a linkage.
	///	<remarks>Extra bytes are used to store the offset vector.
	///	Inform the DependencyLinkManager using the xtrabytes value.</summary>
	explicit OffsetIdDependencyLinkage (Bentley::DgnPlatform::DependencyLinkage const& linkage)
	:	DependencyLinkageBase (linkage, sizeof (DVec3d) /* extra bytes */)
	{
   	}
   	OffsetIdDependencyLinkage () : DependencyLinkageBase (sizeof (DVec3d) /* extra bytes */) {}

};

using OffsetIdDependencyLinkageR	= OffsetIdDependencyLinkage&;
using OffsetIdDependencyLinkageCR	= OffsetIdDependencyLinkage const&;
///////////////////////////////////////////////////////////////////////
///	<summary>Helper classes for our DependencyLinkages.
///	<remarks>OffsetDependencyLinkage  is a specialisation of DependencyLinkageBase for TYPE_XYOFFSET.
///	<para>It has additional functionality to store the DVec3d offset vector.</para>
///	<para>DVec3d occupies unused root ID storage space, since this class needs only one root ID.</para>
///	</remarks></summary>
struct OffsetAssocDependencyLinkage :  public DependencyLinkageBase<OffsetIdDependencyLinkage, DEPENDENCYAPPID_TestDepend, TYPE_XYOFFSET_ASSOC, DEPENDENCY_DATA_TYPE_ASSOC_POINT, DEPENDENCY_ON_COPY_DropDependency>
										, public IDependencyOffsetLinkage
{
	///	<summary>GetOffset pointer from a dependency linkage.
	///	<remarks>static</remarks></summary>
	static DVec3d* GetOffset (DependencyLinkageCR			linkage)
	{
		return (DVec3d*)&linkage.root.assoc[1];
	}
	///	<summary>GetOffset from this object.</summary>
	///	<returns>Get the offset vector.</returns>
	virtual DVec3d const& 	IDependencyOffsetLinkage::GetOffset	()	const
	{
		return *GetOffset (*this);
	}
	///	<summary>ReplaceInModel updates an existing element with the current state of this linkage.</summary>
	///	<remarks>Forwards to base class implementation.</remarks>
	void			ReplaceInModel (ElementHandleCR	eh) const override
	{
		ReplaceInModelImpl (eh);
	}
	///	<summary>SetOffset.</summary>
	///	<returns>Set the offset vector.</returns>
	virtual void			IDependencyOffsetLinkage::SetOffset		(DVec3d const&	offset) override
	{
		//	Nasty C++ hack
		DVec3d* target	= reinterpret_cast<DVec3d*>(&root.assoc[1]);
		*target = offset;
	}
	///	<summary>Copy constructor taking a linkage.
	///	<remarks>Extra bytes are used to store the offset vector.
	///	Inform the DependencyLinkManager using the xtrabytes value.</summary>
	explicit OffsetAssocDependencyLinkage (Bentley::DgnPlatform::DependencyLinkage const& linkage)
	:	DependencyLinkageBase (linkage, sizeof (DVec3d) /* extra bytes */)
	{
   	}
   	OffsetAssocDependencyLinkage () : DependencyLinkageBase (sizeof (DVec3d) /* extra bytes */) {}
};

using OffsetAssocDependencyLinkageR		= OffsetAssocDependencyLinkage&;
using OffsetAssocDependencyLinkageCR	= OffsetAssocDependencyLinkage const&;
///////////////////////////////////////////////////////////////////////
///	<summary>BoxDependencyLinkage.
///	<remarks>BoxDependencyLinkage is a specialisation of DependencyLinkageBase for TYPE_BOX,
///	but has no additional functionality.
///	<para>This is the simplest possible linkage, because it stores no additional app-specific data.</para>
///	</remarks></summary>
struct BoxDependencyLinkage :  public DependencyLinkageBase<BoxDependencyLinkage, DEPENDENCYAPPID_TestDepend, TYPE_BOX, DEPENDENCY_DATA_TYPE_ELEM_ID, DEPENDENCY_ON_COPY_DropDependency>
{
	///	<summary>Copy constructor taking a linkage.</summary>
	explicit BoxDependencyLinkage (Bentley::DgnPlatform::DependencyLinkage const& linkage)
	:	DependencyLinkageBase (linkage /* no extra bytes */)
	{
   	}
   	BoxDependencyLinkage () {}
};

using BoxDependencyLinkageR		= BoxDependencyLinkage&;
using BoxDependencyLinkageCR	= BoxDependencyLinkage const&;

#endif	// !defined(BOX_DEPENDENCY_LINKAGE_H_INCLUDED_)
